#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef double (*func)(double z);
#define PI (3.14159265358979323846)

double rate  = 8363;
double ends   = 1.5;
unsigned char *sample;
char file[128]="out.sam";

func func1, func2, func3, func4;

double freq1 = 261.62, amp1  = 1, env1[10][2] =   {
				{   0, 0   },
				{ 0.1, 1.0 },
				{   1, 0.3 },
				{ 1.5, 0   },
				{ 2.1, 0   } };

double freq2 = 19, amp2  = 0.076, env2[10][2] =   {
				{   0, 0   },
				{   1, 1.0 },
				{ 2.1, 0   } };

double freq3 = 0, amp3  = 0.0, env3[10][2];

double freq4 = 0, amp4  = 0.0, env4[10][2];
			
double envelope(double *env, double t)
{
	int	i;
	double	e;

	for (i = 0; i < 10; i++) {
		e = env[i+i+1];
		if (env[i+i+2] >= t) break;
	}
	e += (t-env[i+i])*(env[i+i+3]-e)/(env[i+i+2]-env[i+i]);
	return e;
}

double fsine(double q)    { return sin(q * 2 * PI); }
double fcosine(double q)  { return cos(q * 2 * PI); }
double fabssine(double q) { return fabs(sin(q * 2 * PI)); }
double fsquare(double q)  { q = q - floor(q); if (q < 0) q = 1 + q; q = q * 2 - 1; return (q < 0) ? -1 : 1; }
double fsaw (double q)    { q = q - floor(q); if (q < 0) q = 1 + q; return q * 2 - 1; }
double ftri(double q)     { q = q - floor(q); if (q < 0) q = 1 + q; return fabs(q * 2 - 1)*2 - 1; }
double fnoise(double q)   { return (rand()-rand())/32767.0; }
double fone(double q)   { return 1; }
double fneg(double q)   { return -1; }

int main(int c, char **v)
{
	FILE	*fh;
	long	i, f;
	double	t, env, x;
	
	func1 = func2 = func3 = func4 = fsine;
	if(c<2) {
		if (!makeparam("fm.dat"))
			printf("4OP linear FM syntheser by Helios Acidum 1999.\n"
				"Usage: %s [<filename>]",v[0]);
		}
	else makeparam(v[1]);

	f = rate * ends;
	sample = (unsigned char *) malloc (f);

	for (i = 0; i < f; i++) {
		t = i / rate;
		env = envelope(env4,t);
		x   = freq3 * env * amp4 * (*func4) (t * freq4);
		env = envelope(env3,t);
		x   = freq2 * env * amp3 * (*func3) (t * (freq3 + x));
		env = envelope(env2,t);
		x   = freq1 * env * amp2 * (*func2) (t * (freq2 + x));
		env = envelope(env1,t);
		x   = 127.0 * env * amp1 * (*func1) (t * (freq1 + x));
		sample[i] = 128.0 + x;
	}

	fh = fopen(file,"wb");
	fwrite(sample,1,i,fh);
	fclose(fh);

	return 0;
}

func getfunc(char *nm)
{
	func	z;
	z  = fsine;
	if (!strcmp(nm,"sin")) z = fsine;
	if (!strcmp(nm,"cos")) z = fcosine;
	if (!strcmp(nm,"abs")) z = fabssine;
	if (!strcmp(nm,"sqr")) z = fsquare;
	if (!strcmp(nm,"saw")) z = fsaw;
	if (!strcmp(nm,"tri")) z = ftri;
	if (!strcmp(nm,"noise")) z = fnoise;
	if (!strcmp(nm,"one")) z = fone;
	if (!strcmp(nm,"neg")) z = fneg;
	return z;
}

char s[6][128];
int makeparam(char *fi)
{
	int	i,ie1,ie2,ie3,ie4;
	double	t;
	FILE	*fh = fopen(fi,"rt");
	if (fh == NULL) return 0;
	fgets(s[0],127,fh);
	sscanf(s[0],"%s%s%s",s[1],s[2],s[3]); /* rate, total, filename */
//	printf("[%s] [%s] [%s] [%s] [%s]\n",s[1],s[2],s[3],s[4],s[5]);
	if (strcmp(s[1],"*")) rate = atof(s[1]);
	if (strcmp(s[2],"*")) ends = atof(s[2]);
	if (strcmp(s[3],"*")) strcpy(file,s[3]);
	/* get main parameres */
	fgets(s[0],127,fh);
	sscanf(s[0],"%s%s%s",s[1],s[2],s[3]); /* freq, amplify, type */
	if (strcmp(s[1],"*")) freq1 = atof(s[1]);
	if (strcmp(s[2],"*")) amp1  = atof(s[2]);
	if (strcmp(s[3],"*")) func1 = getfunc(s[3]);
	fgets(s[0],127,fh);
	sscanf(s[0],"%s%s%s",s[1],s[2],s[3]); /* freq, amplify, type */
	if (strcmp(s[1],"*")) freq2 = atof(s[1]);
	if (strcmp(s[2],"*")) amp2  = atof(s[2]);
	if (strcmp(s[3],"*")) func2 = getfunc(s[3]);
	fgets(s[0],127,fh);
	sscanf(s[0],"%s%s%s",s[1],s[2],s[3]); /* freq, amplify, type */
	if (strcmp(s[1],"*")) freq3 = atof(s[1]);
	if (strcmp(s[2],"*")) amp3  = atof(s[2]);
	if (strcmp(s[3],"*")) func3 = getfunc(s[3]);
	fgets(s[0],127,fh);
	sscanf(s[0],"%s%s%s",s[1],s[2],s[3]); /* freq, amplify, type */
	if (strcmp(s[1],"*")) freq4 = atof(s[1]);
	if (strcmp(s[2],"*")) amp4  = atof(s[2]);
	if (strcmp(s[3],"*")) func4 = getfunc(s[3]);

	for (t=ie1=ie2=ie3=ie4=i=0; i < 9;i++) {
		if (fgets(s[0],127,fh) == NULL) break;
		if (5 != sscanf(s[0],"%s%s%s%s%s",s[1],s[2],s[3],s[4],s[5])) break; /* t, e1, e2, e3 ,e4 */
		if (strcmp(s[1],"*")) t = atof(s[1]);
		if (strcmp(s[2],"*")) { env1[ie1][0] = t; env1[ie1][1] = atof(s[2]); ie1++; }
		if (strcmp(s[3],"*")) { env2[ie2][0] = t; env2[ie2][1] = atof(s[3]); ie2++; }
		if (strcmp(s[4],"*")) { env3[ie3][0] = t; env3[ie3][1] = atof(s[4]); ie3++; }
		if (strcmp(s[5],"*")) { env4[ie4][0] = t; env4[ie4][1] = atof(s[5]); ie4++; }
		}
	env1[ie1][0] = ends+1; env1[ie1][1] = 0;
	env2[ie2][0] = ends+1; env2[ie2][1] = 0;
	env3[ie3][0] = ends+1; env3[ie3][1] = 0;
	env4[ie4][0] = ends+1; env4[ie4][1] = 0;

	fclose(fh);
	return 1;
	}
